// Copyright 1987-2012 by Jon Dart.  All Rights Reserved.

#include "search.h"
#include "globals.h"
#include "notation.h"
#include "movegen.h"
#include "util.h"
#include "hash.h"
#include "see.h"
#ifdef GAVIOTA_TBS
#include "gtb.h"
#endif
#ifdef NALIMOV_TBS
#include "tbprobe.h"
#endif
#include "legal.h"
#include "history.h"
#ifndef _WIN32
#include <errno.h>
#endif

#include <cmath>
#include <iomanip>
#include <list>
#include <vector>

#ifdef UCI_LOG
#include <fstream>
extern fstream ucilog;
#endif

static const int THREAD_SPLIT_MIN_DEPTH = 6;
static const int QSEARCH_FORWARD_PRUNE_BIAS = 5*PAWN_VALUE/4;
static const int ASPIRATION_WINDOW[] =
    {(int)(0.75*PAWN_VALUE), 
     (int)(1.5*PAWN_VALUE),
      3*PAWN_VALUE, 
      PAWN_VALUE*6,
      Constants::MATE};
static const int ASPIRATION_WINDOW_STEPS = 5;

#define VERIFY_NULL_SEARCH
#define STATIC_NULL_PRUNING
#define RAZORING
//#define REFINE_EVAL

static const int FUTILITY_DEPTH = 3*DEPTH_INCREMENT;
static const int CHECK_EXTENSION = 3*DEPTH_INCREMENT/4;
static const int FORCED_EXTENSION = DEPTH_INCREMENT;
static const int PAWN_PUSH_EXTENSION = DEPTH_INCREMENT;
static const int CAPTURE_EXTENSION = DEPTH_INCREMENT/2;
static const int THREAT_EXTENSION = DEPTH_INCREMENT/2;
static const int RAZOR_DEPTH = 3*DEPTH_INCREMENT;
static const int LMR_DEPTH = int(2.25F*DEPTH_INCREMENT);
static const int EASY_THRESHOLD = 2*PAWN_VALUE;
static const int EASY_PLIES = 3;
static const int EXTRA_LMR_START_PLY = 8;
static const int BASE_LMR_REDUCTION = DEPTH_INCREMENT;
static int CACHE_ALIGN LMR_REDUCTION[2][64][64];

static const int FUTILITY_MARGIN_BASE[16] =
    {(int)(1.5*PAWN_VALUE),
     (int)(1.5*PAWN_VALUE),
     (int)(1.5*PAWN_VALUE),
     (int)(1.75*PAWN_VALUE),
     (int)(2.0*PAWN_VALUE),
     (int)(2.25*PAWN_VALUE),
     (int)(2.5*PAWN_VALUE),
     (int)(2.75*PAWN_VALUE),
     (int)(3.0*PAWN_VALUE),
     (int)(3.5*PAWN_VALUE),
     (int)(4.0*PAWN_VALUE),
     (int)(4.5*PAWN_VALUE),
     (int)(5.0*PAWN_VALUE),
     (int)(10.0*PAWN_VALUE),
     (int)(10.0*PAWN_VALUE),
     (int)(10.0*PAWN_VALUE)
};

static int FUTILITY_MARGIN[16][64];

static const int HISTORY_PRUNE_DEPTH = 12;
struct HistoryPruneParams {
  int historyMinMoveCount;
  int evalThreshold;
  int pruningMinMoveCount;
};
static HistoryPruneParams HISTORY_PRUNE_PARAMS[HISTORY_PRUNE_DEPTH*DEPTH_INCREMENT];

static int Time_Check_Interval = 10000;

static int time_check_counter = 0;
static unsigned last_time = 0;

static const int Illegal = Scoring::INVALID_SCORE;
static const int PRUNE = -Constants::MATE;

static int THREAD_SPLIT_DEPTH(const Board &board) {
  return DEPTH_INCREMENT*(THREAD_SPLIT_MIN_DEPTH +
                          (32-board.getMaterial(board.sideToMove()).materialLevel())/16);
}

static FORCEINLINE int calcGain(const Board &board, Move move) {
    int gain = Gain(move);
    if (Capture(move) == Pawn && Rank(DestSquare(move),board.oppositeSide()) == 7) {
        gain += 50;
    }
    if (PieceMoved(move) == Pawn && Rank(DestSquare(move),board.sideToMove()) == 7) {             
        gain += 50;
    }
    return gain;
}

static void setCheckStatus(Board &board, CheckStatusType s)
{
   if (s != CheckUnknown) {
      // Here we take advantage of the fact that if we know the
      // check status before the move, we can set it in the board
      // structure after the move and avoid another calculation.
      board.setCheckStatus(s);
   }
}

SearchController::SearchController()
  : post_function(NULL),
    terminate_function(NULL),
    age(1),
    talkLevel(Silent),
    ratingDiff(0),
    ratingFactor(0),
    active(false) {

#ifdef _WIN32
    HANDLE inh = GetStdHandle(STD_INPUT_HANDLE);
    DWORD dw;
    pipe = !GetConsoleMode(inh, &dw);
    if (!pipe) {
        SetConsoleMode(inh, dw & ~(ENABLE_LINE_INPUT |
            ENABLE_MOUSE_INPUT | ENABLE_WINDOW_INPUT));
    }
#endif
    pool = new ThreadPool(this,options.search.ncpus);
    ThreadInfo *ti = pool->mainThread();
    ti->state = ThreadInfo::Working;
    rootSearch = (RootSearch*)ti->work;
    ti->parentNode = ti->work->getNode();
    for (int d = 0; d < 64; d++) {
      int base = 0;
      if (d>= EXTRA_LMR_START_PLY) {
         base += (int)(log((double)d-EXTRA_LMR_START_PLY+1) * DEPTH_INCREMENT)/4;
      }
      for (int moves= 0; moves < 64; moves++) {
        LMR_REDUCTION[0][d][moves] = 
            LMR_REDUCTION[1][d][moves] = BASE_LMR_REDUCTION;
        int extend = base;
        int steps = Util::Min(4,4*base/DEPTH_INCREMENT);
        int step_size = Util::Max(DEPTH_INCREMENT/4,base/4);
        int d2 = (d<EXTRA_LMR_START_PLY) ? 0 : d/2;
        if (moves > (64-d2) && steps>=1) {
           extend += steps*step_size;
        }
        else if (moves > (32-d2) && steps>=2) {
           extend += (steps-1)*step_size;
        }
        else if (moves > (16-d2) && steps>=3) {
           extend += (steps-2)*step_size;
        }
        else if (moves > (8-d2) && steps>=4) {
           extend += (steps-3)*step_size;
        }
        LMR_REDUCTION[0][d][moves] += extend;
        LMR_REDUCTION[1][d][moves] += 2*extend;
      }
    }
    /*
      for (int i = 8; i < 64; i++) {
      cout << "--- i=" << i << endl;
      for (int m=0; m<64; m++) {
      cout << m << " " << LMR_REDUCTION[0][i][m] << ' ' << LMR_REDUCTION[1][i][m] << endl;
      }
      }
    */
    for (int depth = 0; depth < HISTORY_PRUNE_DEPTH*DEPTH_INCREMENT; depth++) {
        HistoryPruneParams &p = HISTORY_PRUNE_PARAMS[depth];
        p.pruningMinMoveCount = 3 + int((0.5F*depth*depth)/(DEPTH_INCREMENT*DEPTH_INCREMENT));
        p.historyMinMoveCount = 1 + depth/DEPTH_INCREMENT;
        const float depthDivisor = log(2+pow((float)depth/DEPTH_INCREMENT,2.0F));
        p.evalThreshold = (depth <= DEPTH_INCREMENT) ? 
            (int)(0.7F*Constants::HISTORY_MAX) : (int)((0.7F / pow((float)depthDivisor,1.25F))*Constants::HISTORY_MAX);
        /*
        if (depth % DEPTH_INCREMENT == 0) cout << "-- ply " << depth/DEPTH_INCREMENT << endl;
        cout << p.pruningMinMoveCount << ' ' << 
            p.historyMinMoveCount << ' ' <<
            p.evalThreshold << endl;
        */
    }
    for (int i = 0; i < 16; i++) {
        for (int j = 0; j < 64; j++) {
            int fmargin = FUTILITY_MARGIN_BASE[i];
            if (j > 8 && i >= 4) {
               fmargin -= Util::Min(fmargin/4,fmargin*(j-8)/128);
            }
            FUTILITY_MARGIN[i][j] = fmargin;
        }
    }
}

SearchController::~SearchController() {
   delete pool;
}

void SearchController::terminateNow() {
    if (talkLevel == Trace)
        cout << "# terminating search (controller)" << endl;
    rootSearch->stopAllThreads();
}

Move SearchController::findBestMove(
   const Board &board,
   SearchType srcType,
   int time_limit,
   int xtra_time,
   int ply_limit,
   int background,
   int isUCI,
   Statistics &stat_buf,
   TalkLevel t)
{
    active = true;
    explicit_excludes = 0;
    Move excludes[Constants::MaxMoves];
    Move result = findBestMove(board,srcType,time_limit,xtra_time,ply_limit,
                        background, isUCI, stat_buf, t, excludes, 0);
    active = false;
    return result;
}


Move SearchController::findBestMove(
   const Board &board,
   SearchType srcType,
   int time_limit,
   int xtra_time,
   int ply_limit,
   int background,
   int isUCI,
   Statistics &stat_buf,
   TalkLevel t,
   Move *exclude, int num_exclude)
{
    this->typeOfSearch = srcType;
    this->time_limit = time_target = time_limit;
    time_added = 0;
    this->xtra_time = xtra_time;
    if (srcType == FixedTime || srcType == TimeLimit)
        ply_limit = Constants::MaxPly;
    else {
        ply_limit = Util::Min(ply_limit,Constants::MaxPly-1);
        ply_limit = Util::Max(1,ply_limit);
    }
    this->ply_limit = ply_limit;
    this->background = background;
    this->uci = isUCI;
    this->talkLevel = t;
    this->stats = &stat_buf;
    computerSide = board.sideToMove();
    stats->clear();
    explicit_excludes = 0;
    num_exclude = Util::Min(MAX_PV,num_exclude);
    if (num_exclude) explicit_excludes = 1;

    // Positions are stored in the hashtable with an "age" to identify
    // which search they came from. "Newer" positions can replace
    // "older" ones. Update the age here since we are starting a
    // new search.
    age = (age + 1) % 256;

    rootSearch->init(board);
    startTime = getTime();
    return rootSearch->ply0_search(exclude,num_exclude);
}

void SearchController::setRatingDiff(int rdiff)
{
    ratingDiff = rdiff;
    ratingFactor = 0;
    if (ratingDiff < -100) {
        ratingFactor = Util::Max((PAWN_VALUE*(ratingDiff+100))/300,-PAWN_VALUE);
        if (ratingDiff < -400)
          ratingFactor += (400+ratingDiff)*PAWN_VALUE/1000;
    }
    else if (ratingDiff > 100) {
        ratingFactor = Util::Min((PAWN_VALUE*(ratingDiff-100))/300,PAWN_VALUE);
        if (ratingDiff > 400)
          ratingFactor += (ratingDiff-400)*PAWN_VALUE/1000;
    }
}

int SearchController::drawScore(const Board & board) const {
    int score = 0;

    // if we know the opponent's rating (which will be the case if playing
    // on ICC), factor that into the draw score - a draw against a high-rated
    // opponent is good; a draw against a lower-rated one is bad.
    if (ratingDiff != 0) {
        if (board.sideToMove() == computerSide)
           score += ratingFactor;
        else
           score -= ratingFactor;
    }
    return score;
}

int SearchController::wasTerminated() const {
   return rootSearch->wasTerminated();
}

void SearchController::setThreadCount(int threads) {
   pool->resize(threads,this);
}

int SearchController::getIterationDepth() const {
   return rootSearch->getIterationDepth();
}


void SearchController::updateStats(NodeInfo *node, int iteration_depth,
int score, int alpha, int beta)
{
    unsigned end_time = getTime();
    stats->elapsed_time = end_time - startTime;
    stats->multipv_count = rootSearch->multipv_count;
    ASSERT(stats->multipv_count >= 0 && stats->multipv_count < MAX_PV);
    stats->value = score;
    stats->depth = iteration_depth;
    // if failing low, keep the current value for display purposes,
    // not the bottom of the window
    if (stats->value > alpha) {
       stats->display_value = stats->value;
    }
    //stats->state = state;
    Board board_copy(rootSearch->getInitialBoard());

    // note: retain previous best line if we do not have one here
    if (IsNull(node->pv[0])) {
#ifdef _TRACE
        cout << "# warning: pv is null\n";
#endif
        return;
    }
    else if (node->pv_length == 0) {
        return;
    }
    node->best = node->pv[0];                     // ensure "best" is non-null
    ASSERT(!IsNull(node->best));
    stats->best_line[0] = NullMove;
    int i = 0;
    stats->best_line_image.clear();
#ifdef _TRACE
    cout << "best line:" << endl;
#endif
    stringstream best_line_image;
    while (i < node->pv_length && !IsNull(node->pv[i])) {
        ASSERT(i<Constants::MaxPly);
        stats->best_line[i] = node->pv[i];
#ifdef _TRACE
        MoveImage(node->pv[i],cout); cout << ' ' << (flush);
#endif
        ASSERT(legalMove(board_copy,stats->best_line[i]));
        if (i!=0) best_line_image << ' ';
        if (uci) {
            // UCI format for moves
            Notation::UCIMoveImage(stats->best_line[i],best_line_image);
        }
        else {
            // use SAN
            Notation::image(board_copy,stats->best_line[i],best_line_image);
        }
        int len = (int)best_line_image.tellg();
        // limit the length
        if (len > 250) {
            break;
        }
        board_copy.doMove(stats->best_line[i]);
        ++i;
        int rep_count;
        if (Scoring::isDraw(board_copy,rep_count,0)) {
            break;
        }
        if (node->pv_length < 2 && options.search.hash_table_size) {
            // get the next move from the hash table, if possible
            // (for pondering)
            PositionInfo entry;
            PositionInfo::ValueType result =
                Hash::searchHash(board_copy,board_copy.hashCode(rep_count),
                           0,0,0,age,
                           iteration_depth,entry);
            if (result != PositionInfo::NoHit) {
                Move hashMove = entry.bestMove(board_copy);
                if (!IsNull(hashMove)) {
                    stats->best_line[i] = hashMove;
                    if (i!=0) best_line_image << ' ';
                    if (uci) {
                        // UCI format for moves
                        Notation::UCIMoveImage(hashMove,best_line_image);
                    }
                    else {
                        // use SAN
                        Notation::image(board_copy,hashMove,best_line_image);
                    }
                    ++i;
                }
                break;
            }
        }
    }
#ifdef _TRACE
    cout << endl;
#endif
    stats->best_line[i] = NullMove;
    stats->best_line_image = best_line_image.str();
#ifdef _TRACE
    cout << "best line image: " << stats->best_line_image << endl;
#endif
}

void SearchController::clearHashTables()
{
    age = 0;
    pool->clearHashTables();
    Hash::clearHash();
}

void SearchController::uciSendInfos(Move move, int move_index, int depth) {
   if (uci) {
      cout << "info depth " << depth;
      cout << " currmove ";
      Notation::UCIMoveImage(move,cout);
      cout << " currmovenumber " << move_index;
      if (stats->elapsed_time >= 30) {
        cout << " nps " << (100L*stats->num_nodes)/stats->elapsed_time;
      } 
      cout << " nodes " << stats->num_nodes;
      cout << " hashfull " << Hash::pctFull();
      cout << endl;
   }
} 

Search::Search(SearchController *c, ThreadInfo *threadInfo)
    :controller(c),terminate(0),ti(threadInfo),waitTime(0) {
    node = stack;
    LockInit(splitLock);
}

void Search::init(const Board &board) {
  this->board = initialBoard = board;
  context.clearKiller();
}

Search::~Search() {
    LockDestroy(splitLock);
}

int Search::checkTime(const Board &board,int ply) {
    if (terminate) {
       if (controller->talkLevel==Trace) cout << "# check time, already terminated" << endl;
       return 1; // already stopped search
    }

    Statistics *stats = controller->stats;
#if defined(SMP_STATS)
    stats->samples++;
    stats->threads += controller->pool->activeCount();
#endif
    unsigned current_time = getTime();
    if (controller->typeOfSearch != FixedDepth) {
        // update at least the part of the stats structure that
        // is displayed in analysis mode.
        stats->elapsed_time = current_time - controller->startTime;
        if (stats->elapsed_time > controller->getTimeLimit()) {
            if (controller->talkLevel == Trace) {
                cout << "# terminating, time up" << endl;
            }
            return 1;
        }
        if (controller->typeOfSearch == FixedTime) {
            return 0;
        }
        unsigned n = theLog->current();

        if (controller->xtra_time > 0 &&
            controller->typeOfSearch == TimeLimit &&
            (stats->elapsed_time > (unsigned)controller->time_target/3) &&
            controller->time_added == 0) {
            // see if our score has dropped a lot since our
            // last search
            if (n >= 2 && !(*theLog)[n-2].getBookInfo().isValid() &&
                (stats->value < 5*PAWN_VALUE) &&
                (stats->value > -PAWN_VALUE*5) &&
                (stats->value < (*theLog)[n-2].score() - PAWN_VALUE/3)) {
                // search more time because our score is dropping
                controller->time_added++;
                if (controller->talkLevel == Trace) {
                    cout << "# adding time" << endl;
                }
            }
        }
    }
    if (controller->uci && (current_time-last_time >= 200)) {
        cout << "info";
        if (stats->elapsed_time>30) cout << " nps " <<
                (long)((100L*stats->num_nodes)/stats->elapsed_time);
        cout << " nodes " << stats->num_nodes << " hashfull " << Hash::pctFull() << endl;
        last_time = current_time;
    }
    return 0;
}

void Search::showStatus(const Board &board, Move best,int faillow,
int failhigh,int complete)
{
    if (terminate)
        return;
    Statistics *stats = controller->stats;
    int ply = stats->depth;
    stats->complete = complete;
    if (controller->talkLevel == Debug) {
        cout.setf(ios::fixed);
        cout << setprecision(2);
        cout << ply << '\t';
        cout << stats->elapsed_time/100.0 << '\t';
        if (faillow) {
            cout << " --";
        }
        else if (best != NullMove) {
            Notation::image(board, best, cout);
            if (failhigh) cout << '!';
        }
        cout << '\t';
        Scoring::printScore(stats->value,cout);
        cout << '\t' << stats->num_nodes << endl;
    }
    // Post during ponder if UCI
    if ((!controller->background || controller->uci)) {
        if (controller->post_function) {
            controller->post_function(*stats);
        }
    }
}

Move RootSearch::ply0_search(
Move *excludes, int num_excludes)
{
   this->easy_adjust = false;
   last_score = -Constants::MATE;
   node = stack;                                  // top of stack
#ifdef EVAL_STATS
   Scoring::clearStats();
#endif
   node->best = NullMove;
   if (scoring.isLegalDraw(board)) {
      // If it's a legal draw situation before we even move, then
      // just return a draw score and don't search.
      controller->stats->state = Draw;
      controller->stats->value = drawScore(board);
      return NullMove;
   }
   // Generate the ply 0 moves here:
   RootMoveGenerator mg(board,&context,NullMove,
      controller->talkLevel == Trace);
   controller->stats->num_moves += mg.moveCount();

   terminate = 0;
   time_check_counter = 0;
   last_time = 0;

   int tb_hit = 0, tb_pieces = 0;
   int value = Scoring::INVALID_SCORE;
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
   if (options.search.use_tablebases) {
      const Material &wMat = board.getMaterial(White);
      const Material &bMat = board.getMaterial(Black);
      int tb_score;
      tb_pieces = wMat.men() + bMat.men();
      if(tb_pieces <= EGTBMenCount) {
         controller->stats->tb_probes++;
#ifdef NALIMOV_TBS
         if (options.search.tablebase_type == Options::NALIMOV_TYPE) {
             tb_hit = NalimovTb::probe_tb(board, tb_score, 0);
         }
#endif
#ifdef GAVIOTA_TBS
         if (options.search.tablebase_type == Options::GAVIOTA_TYPE) {
             tb_hit = GaviotaTb::probe_tb(board, tb_score, 0, true);
         }
#endif
         if (tb_hit) {
            value = tb_score;
            controller->stats->tb_hits++;
            if (controller->talkLevel == Trace) {
               cout << "# tb hit, score=";
               Scoring::printScore(value,cout);
               cout << endl;
            }
         }
      }
   }
#endif
   easyMove = NullMove;
   if (value == Scoring::INVALID_SCORE) {
      value = 0;
   }
   waitTime = 0;
   // Reduce strength but not in analysis mode:
   if (options.search.strength < 100 && controller->time_target != INFINITE_TIME) {
       int mgCount = mg.moveCount();
       if (mgCount) {
           float factor = (100-options.search.strength)*1.3F;
           if (factor > 100.0F) factor = 100.0F;
           int max = (10*controller->time_target)/(3*mgCount);
           // wait time is in milliseconds
           waitTime = int((max*factor)/100.0F);
           if (controller->talkLevel == Trace) {
               cout << "# waitTime=" << waitTime << endl;
           }
           // adjust time check interval since we are lowering nps
           Time_Check_Interval /= (1+8*int(factor));
           if (options.search.strength <= 90) {
               static const int limits[10] = {1,1,1,1,2,2,3,5,8,12};
               controller->ply_limit = Util::Min(limits[options.search.strength/10],
                                                 controller->ply_limit);
               if (controller->talkLevel == Trace) {
                   cout << "# ply limit =" << controller->ply_limit << endl;
               }
           }
       }
   }

   controller->stats->value =
      controller->stats->display_value = value;

   // Incrementally search the board to greater depths - stop when
   // ply limit, time limit, interrupt, or a terminating condition
   // is reached.
   node->best = node->pv[0] = NullMove;
   node->threat = 0;
   depth_at_pv_change = 0;
   // search the first few iterations with a wide window - for easy
   // move detection.
   wideWindowDepth = EASY_PLIES;
   wideWindow = EASY_THRESHOLD;
   for (iteration_depth = 1;
        iteration_depth <= controller->ply_limit && !terminate;
        iteration_depth++) {
       if (!controller->explicit_excludes) num_excludes = 0;
       for (multipv_count=0; multipv_count < options.search.multipv && !terminate; multipv_count++) {
           int lo_window, hi_window;
           int aspirationWindow = ASPIRATION_WINDOW[0];
           if (iteration_depth <= 1) {
               lo_window = -Constants::MATE;
               hi_window = Constants::MATE;
           } else if (iteration_depth <= EASY_PLIES) {
               lo_window = Util::Max(-Constants::MATE,value - EASY_THRESHOLD);
               hi_window = Util::Min(Constants::MATE,value + aspirationWindow/2);
           } else {
               lo_window = Util::Max(-Constants::MATE,value - aspirationWindow/2);
               hi_window = Util::Min(Constants::MATE,value + aspirationWindow/2);
           }
           if (controller->talkLevel == Trace && controller->background) {
               cout << "# " << iteration_depth << ". move=";
               MoveImage(node->best,cout); cout << " score=" << node->best_score
                                                << " terminate=" << terminate << endl;
           }
           bool failLow = true;
           int fails = 0;
           while (!terminate && failLow) {          // fail low loop

               failLow = false;
               bool failHigh = true;

               while (!terminate && failHigh) {      // fail high loop
                   failHigh = false;
#ifdef _TRACE
                   cout << "iteration " << iteration_depth << " window = [" <<
                       lo_window << "," << hi_window << "]" << endl;
#endif
                   value = ply0_search(mg, lo_window, hi_window, iteration_depth,
                                       DEPTH_INCREMENT*iteration_depth,
                                       excludes,num_excludes);
#ifdef _TRACE
                   cout << "iteration " << iteration_depth << " result: " <<
                       value << endl;
#endif
                   controller->updateStats(stack,iteration_depth,
                                           value,lo_window,hi_window);
                   if (iteration_depth == 0 && scoring.isLegalDraw(board)) {
                       if (controller->talkLevel == Trace)
                           cout << "# draw, terminating" << endl;
                       controller->terminateNow();
                   }
                   // check for forced move, but only at depth 2
                   // (so we get a ponder move if possible). But exit immediately
                   // if a tb hit because deeper search will hit the q-search and
                   // the score will be inaccurate. Do not terminate here if a
                   // resign score is returned (search deeper to get an accurate
                   // score). Do not exit in analysis mode.
                   else if (!(controller->background || controller->time_target == INFINITE_TIME) &&
                            mg.moveCount() == 1 &&
                            (iteration_depth >= 2) &&
                            (!options.search.can_resign ||
                             (controller->stats->display_value >
                              options.search.resign_threshold))) {
                       if (controller->talkLevel == Trace)
                           cout << "# single legal move, terminating" << endl;
                       controller->terminateNow();
                   }
                   Statistics *stats = controller->stats;
                   StateType &state = stats->state;
                   if (!terminate && (state == Checkmate || state == Stalemate)) {
                       if (controller->talkLevel == Trace)
                           cout << "# terminating due to checkmate or statemate, state="
                                << (int)state << endl;
                       controller->terminateNow();
                       break;
                   }
                   if (stats->elapsed_time > 100) {
                       long denom = controller->background ? 200L : 500L;
                       Time_Check_Interval = (int)(
                                                   (denom*stats->num_nodes)/(36L*stats->elapsed_time));
                       if (controller->talkLevel == Trace) {
                           cout << "# time check interval=" << Time_Check_Interval << " elapsed_time=" << stats->elapsed_time << " target=" << controller->getTimeLimit() << endl;
                       }
                   }
                   if (!terminate) {
                       if (checkTime(board,0)) {
                           if (controller->talkLevel == Trace)
                               cout << "# time up" << endl;
                           controller->terminateNow();
                       }
                       else if (controller->terminate_function &&
                                controller->terminate_function(*stats)) {
                           if (controller->talkLevel == Trace)
                               cout << "# terminating due to program or user input" << endl;
                           controller->terminateNow();
                       }
                   }
                   failHigh = value >= hi_window && (hi_window < Constants::MATE-iteration_depth-1);
                   if (failHigh && !terminate) {
#ifdef _TRACE
                       cout << "ply 0 high cutoff, re-searching ... value=";
                       Scoring::printScore(value,cout);
                       cout << " fails=" << fails << endl;
#endif
                       ASSERT(fails+1<ASPIRATION_WINDOW_STEPS);
                       aspirationWindow = ASPIRATION_WINDOW[++fails];

                       if (aspirationWindow == Constants::MATE) {
                           hi_window = Constants::MATE-iteration_depth-1;
                       } else {
                           hi_window = Util::Min(Constants::MATE-iteration_depth-1,
                                             lo_window + aspirationWindow);
                       }
                   }
               }                                     // fail high loop
               failLow = value <= lo_window && (value > iteration_depth-Constants::MATE);
               if (failLow && !terminate) {
#ifdef _TRACE
                   cout << "ply 0 fail low, re-searching ..." << endl;
                   cout << "fails = " << fails << endl;
#endif
                   // continue loop with lower bound
                   ASSERT(fails+1 < ASPIRATION_WINDOW_STEPS);
                   aspirationWindow = ASPIRATION_WINDOW[++fails];
                   if (aspirationWindow == Constants::MATE) {
                       lo_window = iteration_depth-Constants::MATE;
                   } else {
                       lo_window = Util::Max(iteration_depth-Constants::MATE,hi_window - aspirationWindow);
                   }
                   showStatus(board, node->best, failLow, failHigh, 0);
               }
           }
           if (!MovesEqual(node->best,easyMove)) {
               depth_at_pv_change = iteration_depth;
           }
#ifdef _TRACE
           cout << iteration_depth << " ply search result: ";
           MoveImage(node->best,cout);
           cout << " value = ";
           Scoring::printScore(value,cout);
           cout << endl;
#endif
           last_score = value;
           if (options.search.multipv > 1) {
               ASSERT(num_excludes<Constants::MaxMoves);
               excludes[num_excludes++] = node->pv[0];
           }
           else if (!(controller->uci && controller->time_limit == INFINITE_TIME)) {
               // Allow early termination on a tablebase position, but not
               // if we have >=6 man tbs in use (because tb set may be
               // incomplete - in that case it is better to allow us to
               // search deeper those nodes that don't produce a tb hit).
               //
               if (tb_hit && tb_pieces<6 && iteration_depth>=3 && !IsNull(node->pv[0])) {
                   if (controller->talkLevel == Trace)
                       cout << "# terminating, tablebase hit" << endl;
#ifdef _TRACE
                   cout << "terminating, tablebase hit" << endl;
#endif
                   controller->terminateNow();
                   break;
               }
               else if (!controller->background &&
                        !easy_adjust && depth_at_pv_change <= EASY_PLIES &&
                        MovesEqual(easyMove,node->best) &&
                        !faillows &&
                        (controller->stats->elapsed_time >
                         (unsigned)controller->time_target/3)) {
                   easy_adjust = true;
                   if (controller->talkLevel == Trace) {
                       cout << "# easy move, adjusting time lower" << endl;
                   }
                   controller->time_target /= 3;
               }
               if (value <= iteration_depth - Constants::MATE) {
                   // We're either checkmated or we certainly will be, so
                   // quit searching.
                   if (controller->talkLevel == Trace)
                       cout << "# terminating, low score" << endl;
#ifdef _TRACE
                   cout << "terminating, low score" << endl;
#endif
                   controller->terminateNow();
                   break;
               }
               else if (value >= Constants::MATE - iteration_depth - 1) {
                   // found a forced mate, terminate
                   if (iteration_depth>=2) {
                       if (controller->talkLevel == Trace)
                           cout << "# terminating, mate score" << endl;
#ifdef _TRACE
                       cout << "terminating, mate score" << endl;
#endif
                       controller->terminateNow();
                       break;
                   }
               }
           }
       }
   }

#ifdef UCI_LOG
   ucilog << "out of search loop " << endl << (flush);
#endif
   // search done, set status and report statistics
   static const int end_of_game[] = {0, 1, 0, 1, 1, 1, 1};
   Statistics *stats = controller->stats;
   StateType &state = stats->state;
   stats->end_of_game = end_of_game[(int)stats->state];
   if (!controller->uci && !stats->end_of_game && options.search.can_resign) {
      if (stats->depth >= 2) {
         // Use the display value - i.e. do not use score if
         // we are failing low
         if (stats->display_value <= 8-(int)Constants::MATE) {
            // TBD we can sometimes get -10000 as a score!
            // this is bogus, we should not resign.
            if (stats->display_value != -(int)Constants::MATE)
               state = Resigns;
         }
         else if ((100*stats->display_value)/PAWN_VALUE <= options.search.resign_threshold) {
            state = Resigns;
         }
         stats->end_of_game = end_of_game[(int)state];
      }
   }

   if (controller->talkLevel == Debug) {
      cout.setf(ios::fixed);
      cout << setprecision(2);
      if (stats->elapsed_time > 0) {
         stats->printNPS(cout);
         cout << " nodes/second." << endl;
      }
#ifdef SEARCH_STATS
      cout << (stats->num_nodes-stats->num_qnodes) << " regular nodes, " <<
         stats->num_qnodes << " quiescence nodes." << endl;
      cout << stats->hash_searches << " searches of hash table, " <<
         stats->hash_hits << " successful";
      if (stats->hash_searches != 0)
         cout << " (" <<
            (int)((100.0*(float)stats->hash_hits)/((float)stats->hash_searches)) <<
            " percent).";
      cout << endl;
      cout << "hash table is " << setprecision(2) <<
          1.0F*Hash::pctFull()/10.0F << "% full." << endl;
#endif
#ifdef MOVE_ORDER_STATS
      cout << "move ordering: ";
      static const char *labels[] = {"1st","2nd","3rd","4th"};
      for (int i = 0; i < 4; i++) {
         cout << setprecision(2) << labels[i] << " " <<
            (100.0*stats->move_order[i])/(float)stats->move_order_count << "% " ;
      }
      cout << endl;
#endif
#ifdef SEARCH_STATS
      cout <<  "null cutoffs: " << stats->num_null_cuts;
      if (stats->num_nodes) {
         cout << " (" <<
            (int)((100.0*(float)stats->num_null_cuts)/((float)stats->num_nodes)) <<
            "%)";
      }
      cout << endl;
      cout << "pruning: " <<
         stats->futility_pruning << " moves (futility)";
      cout << ", ";
      cout << stats->num_razored << " nodes razored";
      cout << endl;
      cout << "extensions: " <<
         stats->check_extensions << " check, " <<
         stats->recap_extensions << " recapture, " <<
         stats->pawn_extensions << " pawn, " <<
         stats->threat_extensions << " threat, " <<
         stats->forced_extensions << " forced move." << endl;
#endif
      cout << stats->tb_probes << " tablebase probes, " <<
         stats->tb_hits << " tablebase hits" << endl;
#if defined(SMP_STATS)
      cout << stats->splits << " splits," <<
         " average thread usage=" << (float)(stats->threads)/(float)stats->samples << endl;
#endif
#ifdef EVAL_STATS
      Scoring::showStats(cout);
#endif
      cout << (flush);
   }
#ifdef UCI_LOG
   ucilog << "out of search" << endl << (flush);
#endif
   return node->best;
}

int RootSearch::ply0_search(RootMoveGenerator &mg, int alpha, int beta,
                        int iteration_depth,
int depth, Move exclude [], int num_exclude)
{
    // implements alpha/beta search for the top most ply.  We use
    // the negascout algorithm.

    time_check_counter++;

    controller->stats->num_nodes++;

    int in_pv = 1;
    int in_check = 0;

    const bool wide = iteration_depth <= wideWindowDepth;

    in_check = (board.checkStatus() == InCheck);
    BoardState save_state = board.state;

    int try_score = alpha;
    //
    // re-sort the ply 0 moves.
    mg.reorder(node->best,controller->getIterationDepth());
    // if in N-variation mode, exclude any moves we have searched already
    mg.exclude(exclude,num_exclude);

    if (controller->getIterationDepth() == EASY_PLIES+1) {
        vector<MoveEntry> &list = mg.getMoveList();
        if (list.size() > 1 && list[0].score >= list[1].score + EASY_THRESHOLD) {
            easyMove = node->best;
            if (controller->talkLevel == Trace) {
                cout << "#easy move: ";
                MoveImage(easyMove,cout);
                cout << endl;
            }
        }
    }

    //
    // Search the next ply
    //
    node->parentNode = node;
    node->pv[0] = NullMove;
    node->pv_length = 0;
    node->cutoff = 0;
    node->extensions = 0;
    node->num_try = 0;                            // # of legal moves tried
    node->alpha = alpha;
    node->beta = beta;
    node->best_score = node->alpha;
#ifdef MOVE_ORDER_STATS
    node->best_count = 0;
#endif
    node->fpruned_moves = 0;
    node->ply = 0;
    node->depth = depth;
    node->split = NULL;
    node->splits = 0;
    node->mg = &mg;
    node->futilityEval = Scoring::INVALID_SCORE;
    node->threatMove = NullMove;
    // clear split point stack:
    splitStackEnd = splitStack;

    int move_index = 0;
    int hibound = beta;
    const int canSplit = options.search.ncpus>1 &&
      depth >= THREAD_SPLIT_DEPTH(board);
    while (!node->cutoff && !terminate) {
        Move move;
        if ((move  = mg.nextMove(node))==NullMove) break;
        move_index++;
        if (IsUsed(move)) {
           continue;     // skip move
        }
        node->last_move = move;
        controller->stats->mvleft = controller->stats->mvtot-move_index;
        if (controller->uci && controller->stats->elapsed_time > 30) {
            controller->uciSendInfos(move, move_index, controller->getIterationDepth());
        }
#ifdef _TRACE
        cout << "trying 0. ";
        MoveImage(move,cout);
        cout << " (" << move_index << "/" << mg.moveCount();
        cout << ")" << endl;
#endif
        node->last_move = move;
        node->extensions = 0;
        CheckStatusType in_check_after_move = board.wouldCheck(move);
        int extend = calcExtensions(board,node,node,in_check_after_move,
                                    node->num_try,
                                    move);
        if (extend == PRUNE) {
#ifdef _TRACE
            cout << "fwd pruned." << endl;
#endif
            node->fpruned_moves++;
            continue;
        }
        ply0nodeCount = mg.getNodeCount(move);
        nodeCount = 0ULL;
        board.doMove(move);
        setCheckStatus(board,in_check_after_move);
        node->num_try++;
#ifdef _TRACE
        cout << "window [" << -hibound << ", " << -node->best_score <<
          "]" << endl;
#endif
        int lobound = wide ? node->alpha : node->best_score;
        if (depth+extend-DEPTH_INCREMENT > 0) {
           try_score = -search(-hibound, -lobound,
                               1, depth+extend-DEPTH_INCREMENT);
        }
        else {
           try_score = -quiesce(-hibound, -lobound, 1, 0);
        }
        if (terminate) {
            board.undoMove(move,save_state);
            break;
        }
#ifdef _TRACE
        cout << "0. ";
        MoveImage(move,cout);
        cout << ' ' << try_score;
        if (in_pv) cout << " (pv)";
        cout << endl;
#endif
        if (!in_pv && try_score > node->best_score && !((node+1)->flags & EXACT) && !terminate) {
           // We failed to get a cutoff and must re-search
#ifdef _TRACE
            cout << "window = [" << -hibound << "," << node->best_score
              << "]" << endl;
            cout << "score = " << try_score << " - no cutoff, researching .." << endl;
#endif
            if (extend < 0) {
                extend = 0;
                node->extensions = 0;
            }
            if (try_score > node->best_score) {
              if (depth+extend-DEPTH_INCREMENT > 0)
                try_score=-search(-node->beta,-lobound,1,depth+extend-DEPTH_INCREMENT);
              else
                try_score=-quiesce(-node->beta,-lobound,1,0);
            }
#ifdef _TRACE
            cout << "0. ";
            MoveImage(move,cout);
            cout << ' ' << try_score;
            cout << endl;
#endif
        }
        board.undoMove(move,save_state);
        (*ply0nodeCount) += nodeCount;
        if (waitTime) {
            // we are in reduced strength mode, waste some time
            sleep(waitTime);
        }
        if (terminate || node->cutoff) {
            break;
        }
        if (try_score > node->best_score) {
            if (updateRootMove(board,node,node,move,try_score)) {
                // beta cutoff
                // ensure we send UCI output .. even in case of quick
                // termination due to checkmate or whatever
                controller->uciSendInfos(move, move_index, controller->getIterationDepth());

                break;
            }
        }
        hibound = node->best_score + 1;  // zero-width window
        in_pv = 0;
        if (canSplit && maybeSplit(board,node,0,depth)) {
            break;
        }
    }

    // We may have other threads still searching.
    maybeWait(node);
    if (node->cutoff) {
        return node->best_score;
    }
    if (node->num_try == 0) {
        // no moves were tried
        if (in_check) {
            if (mg.moveCount() == 0) {           // mate
                node->best_score = -(Constants::MATE);
                controller->stats->state = Checkmate;
            }
        }
        else {                                    // stalemate
            controller->stats->state = Stalemate;
#ifdef _TRACE
            cout << "stalemate!" << endl;
#endif
            node->best_score = drawScore(board);
        }
    }
#ifdef MOVE_ORDER_STATS
    if (rootNode->num_try && rootNode->best_score > rootNode->alpha) {
        stats->move_order_count++;
        if (rootNode->best_count<4) {
            stats->move_order[rootNode->best_count]++;
        }
    }
#endif
    ASSERT(node->best_score >= -Constants::MATE && node->best_score <= Constants::MATE);
    return node->best_score;
}

int Search::quiesce(int ply,int depth)
{
    // recursive function, implements quiescence search.
    //
    ASSERT(ply < Constants::MaxPly);
    controller->stats->num_nodes++;
    nodeCount++;
#ifdef SEARCH_STATS
    controller->stats->num_qnodes++;
#endif
    time_check_counter++;
    int rep_count;
    if (ply >= Constants::MaxPly-1) {
        return scoring.evalu8(board,node->alpha,node->beta);
    }
    else if (Scoring::isDraw(board,rep_count,ply)) {
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "draw!" << endl;
        }
#endif
        node->flags |= EXACT;
        return drawScore(board);
    }
#ifdef _TRACE
    if (master()) {
        indent(ply); cout << "window [" << node->alpha << ","
            << node->beta << "]" << endl;
    }
#endif
    if (board.checkStatus((node-1)->last_move)==InCheck) {
        // If last move was a checking move, ensure that in making it we did
        // not move a pinned piece or move the king into check (normally we
        // would detect this by finding the King can be captured, but if in
        // check we only generate evasions and will not find this).
        if (!board.wasLegal((node-1)->last_move)) {
          ASSERT(board.anyAttacks(board.kingSquare(board.oppositeSide()),board.sideToMove()));
           return -Illegal;
        }
        else {
#ifdef _TRACE
          indent(ply); cout << "in_check=1" << endl;
#endif
           return qsearch_check(ply,depth);
        }
    }
    else {
        return qsearch_no_check(ply,depth);
    }
}

void RootSearch::clearHashTables() {
  Search::clearHashTables();
  History::clearHistory();
  context.clearKiller();
  scoring.clearHashTables();
}

static inline void swap( Move moves[], int i, int j)
{
   Move tmp = moves[j];
   moves[j] = moves[i];
   moves[i] = tmp;
}

int Search::qsearch_no_check(int ply, int depth)
{
    int stand_pat_score;
    // Establish a default score.  This score is returned if no
    // captures are generated, or if no captures generate a better
    // score (since we generally can choose whether or not to capture).
    int bitscore;
    if (board.getMaterial(White).materialLevel()==0 &&
        board.getMaterial(Black).materialLevel()==0 &&
    ((bitscore=Scoring::tryBitbase(board))!=Scoring::INVALID_SCORE)) {
        stand_pat_score = bitscore;
    }
    else {
        stand_pat_score = scoring.evalu8(board,node->alpha,node->beta);
    }
#ifdef _TRACE
    if (master()) {
        indent(ply);
        cout << "stand pat score = " << stand_pat_score << endl;
    }
#endif
    node->best_score = node->alpha;
    if (stand_pat_score > node->best_score) {
        if (stand_pat_score >= node->beta) {
#ifdef _TRACE
            if (master()) {
                indent(ply); cout << "**CUTOFF**" << endl;
            }
#endif
            // check legality of prev move but not at depth == 0 (because
            // regular search checked already)
            if (depth < 0 && !board.wasLegal((node-1)->last_move)) {
               ASSERT(board.anyAttacks(board.kingSquare(board.oppositeSide()),board.sideToMove()));
               return -Illegal;
            }
            else
               return stand_pat_score;
        }
        node->best_score = stand_pat_score;
        (node+1)->pv_length=0;
    }
    const int bias = QSEARCH_FORWARD_PRUNE_BIAS + stand_pat_score;

    Move pv = Scoring::getBestMove(board);

    int move_index = 0;
    int try_score;
    BoardState state = board.state;
    if (!IsNull(pv)) {
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "qsearch cache hit: "; MoveImage(pv,cout);
            cout << endl;
        }
#endif
        if (Capture(pv) == King) {
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
        node->last_move = pv;
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "trying " << ply << ". ";
            MoveImage(pv,cout);
            cout << endl;
        }
#endif
        int gain = calcGain(board,pv);
        if (gain + bias < node->best_score) {
            // Futility pruning.
            // Even assuming we gain the captured piece w/o loss,
            // plus a bias factor, we will not do better than the
            // current best score. So prune this move.
#ifdef _TRACE
           if (master()) {
             indent(ply); cout << "fwd pruned." << endl;
           }
#endif
           goto skip;
        }
        // we do not do SEE on the pv move because it already
        // passed that test - or it would have been pruned
        board.doMove(pv);
        try_score = -quiesce(-node->beta, -node->best_score, ply+1, depth-1);
        board.undoMove(pv,state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(pv,cout);
                cout << "(pv) " << try_score << endl;
            }
#endif
            if (try_score > node->best_score) {
                node->best_score = try_score;
                node->best = pv;
                if (try_score >= node->beta)
                    goto search_end;
                if (node->best_score >= Constants::MATE-1-ply)
                    goto search_end;              // mating move found
            }
        }
    }
skip:
    {
        MoveGenerator mg(board, &context, ply,
                         NullMove, master());
        Move *moves = (Move*)node->done;
        const ColorType oside = board.oppositeSide();
        Bitboard targets(board.occupied[oside]);
        // Pre-select the piece(s) whose capture will be considered.
        // Pieces whose capture is unlikely to reach the futility threshold
        // will not be included in move generation.
        if (bias + (int)(1.5*PAWN_VALUE) < node->best_score) {
            targets &= ~(board.pawn_bits[oside] & ~Attacks::rank7mask[oside]);
            if (bias + KNIGHT_VALUE + (int)(5.0*PAWN_VALUE) < node->best_score) {
               targets &= ~(board.knight_bits[oside] | board.bishop_bits[oside]);
               if (bias + ROOK_VALUE + (int)(8.0*PAWN_VALUE) < node->best_score) {
                  targets &= ~board.rook_bits[oside];
               }
            }
        }
        // generate all the capture moves
        int move_count = mg.generateCaptures(moves,targets);
#ifdef SEARCH_STATS
        controller->stats->num_moves += move_count;
#endif
        mg.initialSortCaptures(moves, move_count);
        while (move_index < move_count) {
            Move move = moves[move_index++];
            if (MovesEqual(move,pv)) continue;  // already did this one
            if (Capture(move) == King) {
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "previous move illegal, exiting qsearch" << endl;
                }
#endif
                return -Illegal;
            }
            int gain = calcGain(board,move);
            if (gain + bias < node->best_score) {
              // Futility pruning.
              // Even assuming we gain the captured piece w/o loss,
              // plus a bias factor, we will not do better than the
              // current best score. So prune this move.
              continue;
            }
            if (gain - PieceValue(PieceMoved(move)) < 0 && see(board,move) < 0) {
              // This appears to be a losing capture
              continue;
            }
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "trying " << ply << ". ";
                MoveImage(move,cout);
                cout << endl;
            }
#endif
            node->last_move = move;
            board.doMove(move);
            try_score = -quiesce(-node->beta, -node->best_score, ply+1, depth-1);
            board.undoMove(move,state);
            if (try_score != Illegal) {
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << ply << ". ";
                    MoveImage(move,cout);
                    cout << ' ' << try_score << endl;
                }
#endif
                if (try_score > node->best_score) {
                    node->best_score = try_score;
                    node->best = move;
                    if (try_score >= node->beta) {
#ifdef _TRACE
                       if (master()) {
                          indent(ply);
                          cout << "**CUTOFF**" << endl;
                        }
#endif
                        goto search_end;
                    }
                    if (node->best_score >= Constants::MATE-1-ply)
                        goto search_end;          // mating move found
                }
            }
#ifdef _TRACE
            else if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << " (illegal)" << endl;
            }
#endif
        }
        // Do checks in qsearch
        if ((stand_pat_score >= node->alpha - 2*PAWN_VALUE) &&
            (depth >= 1-options.search.checks_in_qsearch)) {
            MoveGenerator mg(board, &context, ply, NullMove, master());
            move_count = mg.generateChecks(moves);
            move_index = 0;
#ifdef _TRACE
            if (master()) {
                if (move_count) {
                    indent(ply);
                    cout << move_count << " checks generated" << endl;;
                }
            }
#endif
            while (move_index < move_count) {
                Move move = moves[move_index++];
                if (MovesEqual(move,pv)) continue;
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "trying " << ply << ". ";
                    MoveImage(move,cout);
                    cout << endl;
                }
#endif
                // prune checks that cause loss of the checking piece
                if (see(board,move) < 0) {
#ifdef _TRACE
                  if (master()) {
                     indent(ply); cout << "pruned" << endl;
                  }
#endif
                  continue;
                }
                node->last_move = move;
                board.doMove(move);
                // verify opposite side in check:
                ASSERT(board.anyAttacks(board.kingSquare(board.sideToMove()),board.oppositeSide()));
                // and verify quick check confirms it
                ASSERT(board.checkStatus(move)==InCheck);
                try_score = -quiesce(-node->beta,-node->best_score,
                                     ply+1,depth-1);
                board.undoMove(move,state);
                if (try_score != Illegal) {
                    // Move generator does not prevent generation
                    // of moves of pinned pieces that give check.
                    if (board.isPinned(board.sideToMove(),move)) {
                       continue;
                    }
#ifdef _TRACE
                    if (master()) {
                        indent(ply);
                        cout << ply << ". ";
                        MoveImage(move,cout);
                        cout << ' ' << try_score << endl;
                    }
#endif
                    if (try_score > node->best_score) {
                        node->best_score = try_score;
                        node->best = move;
                        if (try_score >= node->beta) {
#ifdef _TRACE
                            indent(ply); cout << "**CUTOFF**" << endl;
#endif
                            goto search_end;
                        }
                        if (node->best_score >= Constants::MATE-1-ply)
                            goto search_end;      // mating move found
                    }
                }
#ifdef _TRACE
                else if (master()) {
                    indent(ply);
                    cout << ply << ". ";
                    MoveImage(move,cout);
                    cout << " (illegal)" << endl;
                }
#endif
            }
        }
    }
    search_end:
    if (!IsNull(node->best)) {
#ifdef _TRACE
      indent(ply); cout << "storing best move ";
      MoveImage(node->best,cout);
      cout << endl;
#endif
      Scoring::cacheBestMove(board,node->best);
    }
    if (node->inBounds(node->best_score)) {
        if (!IsNull(node->best)) {
            updatePV(board,node->best,ply);
        }
    }
    return node->best_score;
}


int Search::qsearch_check(int ply, int depth)
{
    int try_score;
    Move pvmove = Scoring::getBestMove(board);
    if (!IsNull(pvmove)) {
#ifdef _TRACE
       indent(ply); cout << "qsearch cache hit: ";
       MoveImage(pvmove,cout);
       cout << endl;
#endif
      ASSERT(legalMove(board,pvmove));
    }
    MoveGenerator mg(board, &context, ply, pvmove, master());

    Move move;
    BoardState state = board.state;
    node->num_try = 0;
    while ((move = mg.nextEvasion()) != NullMove) {
        ASSERT(OnBoard(StartSquare(move)));
        if (Capture(move) == King) {
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "previous move illegal, exiting qsearch" << endl;
            }
#endif
            return -Illegal;
        }
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "trying " << ply << ". ";
            MoveImage(move,cout);
            cout << endl;
        }
#endif
        node->last_move = move;
        board.doMove(move);
        ASSERT(!board.anyAttacks(board.kingSquare(board.oppositeSide()),board.sideToMove()));
        try_score = -quiesce(-node->beta, -node->best_score, ply+1, depth-1);
        board.undoMove(move,state);
        if (try_score != Illegal) {
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << ' ' << try_score << endl;
            }
#endif
            node->num_try++;
            if (try_score > node->best_score) {
                node->best_score = try_score;
                node->best = move;
                if (try_score >= node->beta)
                    break;
                if (node->best_score >= Constants::MATE-1-ply)
                    break;                        // mating move found
            }
        }
#ifdef _TRACE
        else if (master()) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(move,cout);
            cout << " (illegal)" << endl;
        }
#endif
    }
    if (node->best_score >= node->beta) {
      if (!IsNull(node->best)) {
#ifdef _TRACE
         indent(ply); cout << "storing best move ";
         MoveImage(node->best,cout);
         cout << endl;
#endif
         Scoring::cacheBestMove(board,node->last_move);
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "**CUTOFF**" << endl;
        }
#endif
      }
      return node->best_score;
    }
    if (node->num_try == 0) { // no legal evasions, so this is checkmate
#ifdef _TRACE
        if (master()) {
          indent(ply); cout << "checkmate!" << endl;
        }
#endif
        node->best_score = -(Constants::MATE - ply);
        (node+1)->pv_length=0; // no PV from this point
        node->flags |= EXACT;
    }
    if (node->best != NullMove) {
#ifdef _TRACE
       indent(ply); cout << "storing best move ";
       MoveImage(node->best,cout);
       cout << endl;
#endif
       Scoring::cacheBestMove(board,node->best);
    }
    if (node->inBounds(node->best_score)) {
       if (!IsNull(node->best)) {
          updatePV(board,node->best,ply);
       }
    }
    return node->best_score;
}

static int FORCEINLINE passedPawnPush(const Board &board, Move move) {
    return (PieceMoved(move) == Pawn &&
            Rank(DestSquare(move),board.sideToMove()) == 7);
}

static int FORCEINLINE passedPawnMove(const Board &board, Move move, int rank) {
  extern CACHE_ALIGN Bitboard passedW[64];
  extern CACHE_ALIGN Bitboard passedB[64];
  if (PieceMoved(move) == Pawn && Rank(DestSquare(move),board.sideToMove()) >= rank) {
    if (board.sideToMove() == White) {
        return Bitboard(board.pawn_bits[Black] & passedW[DestSquare(move)]).is_clear();
    }
    else {
        return Bitboard(board.pawn_bits[White] & passedB[DestSquare(move)]).is_clear();
    }
  }
  else
    return 0;
}

int Search::calcExtensions(const Board &board,
                           NodeInfo *node, NodeInfo *parentNode,
                           CheckStatusType in_check_after_move,
                           int moveIndex,
                           Move move) {
   // see if we should apply any extensions at this node.
   int ply = node->ply;
   int depth = node->depth;
   node->extensions = 0;
   int extend = 0;
   int pruneOk = 1;
   if (board.checkStatus() == InCheck) { // evading check
      pruneOk = 0;
      if (IsForced(move)) {
         // one single reply to check
         extend += FORCED_EXTENSION;
         node->extensions |= FORCED;
#ifdef SEARCH_STATS
         controller->stats->forced_extensions++;
#endif
      }
      else if (IsForced2(move)) {
         // two replies to check
         extend += FORCED_EXTENSION / 2;
         node->extensions |= FORCED;
#ifdef SEARCH_STATS
         controller->stats->forced_extensions++;
#endif
      }
   }
   if (in_check_after_move == InCheck) { // move is a checking move
      node->extensions |= CHECK;
      if (node->PV() || see(board,move) >= -PAWN_VALUE) {
         extend += CHECK_EXTENSION;
      }
      else {
         extend += DEPTH_INCREMENT/2;
      }
      pruneOk = 0;
#ifdef SEARCH_STATS
      controller->stats->check_extensions++;
#endif
   }
   if (passedPawnPush(board,move)) {
      node->extensions |= PAWN_PUSH;
      extend += PAWN_PUSH_EXTENSION;
#ifdef SEARCH_STATS
      controller->stats->pawn_extensions++;
#endif
   }
   else if (Capture(move) != Empty && Capture(move) != Pawn &&
            PieceValue(Capture(move)) == PieceValue(PieceMoved(move)) &&
            board.getMaterial(board.sideToMove()).pieceCount() <= 2) {
      // recapture when there are few pieces
      node->extensions |= CAPTURE;
      extend += CAPTURE_EXTENSION;
   }
   if (parentNode->threat) {
      extend += THREAT_EXTENSION;
      node->extensions |= THREAT;
#ifdef SEARCH_STATS
      controller->stats->threat_extensions++;
#endif
   }
   if (extend || !pruneOk) {
      if (ply > 2*root()->getIterationDepth()) {
         extend /= 2;
      }
      return Util::Min(extend,DEPTH_INCREMENT);
   }

    pruneOk &= moveIndex &&
       Capture(move) == Empty &&
       TypeOfMove(move) == Normal &&
       !passedPawnMove(board,move,5) &&
       !Scoring::mateScore(parentNode->alpha) &&
       (board.getMaterial(White).men() +
        board.getMaterial(Black).men() > 6);

    if (pruneOk) {
        const Move &threat = parentNode->threatMove;
        if (!IsNull(threat)) {
            if (StartSquare(move) == DestSquare(threat)) {
                // threatened piece is being moved, don't prune
                pruneOk = 0;
            } else if (Sliding(board[StartSquare(threat)])) {
                Bitboard btwn;
                board.between(StartSquare(threat),DestSquare(threat),btwn);
                if (btwn.isSet(DestSquare(move)) && see(board,move) >= 0) {
                    // interposition
                    pruneOk = 0;
                }
            }
            if (pruneOk) {
                // check for move defending threatened piece (idea from Stockfish)
                PieceType cap = Capture(threat);
                PieceType pm = PieceMoved(threat);
                if (cap != Empty &&
                    (PieceValue(cap) >= PieceValue(pm) || pm == King) &&
                    board.wouldAttack(move,DestSquare(threat))) {
                    pruneOk = 0; // don't prune
                }
            }
        }
    }

    if (!node->PV() && ply > 0 && pruneOk) {
       // futility pruning, enabled at low depths
       if (depth <= FUTILITY_DEPTH) {
          // threshold increases with move index
          int fmargin = FUTILITY_MARGIN[(depth*4)/DEPTH_INCREMENT][Util::Min(moveIndex,63)];
          int threshold = parentNode->beta - fmargin;
          int eval;
          if (parentNode->futilityEval != Scoring::INVALID_SCORE) {
              eval = parentNode->futilityEval;
          } else {
              eval = parentNode->futilityEval = scoring.evalu8(board,threshold-1,threshold);
          }
          if (eval < threshold) {
#ifdef SEARCH_STATS
              controller->stats->futility_pruning++;
#endif
              return PRUNE;
          }
       }
       // History pruning. Prune scores with low history values.
       if(depth < HISTORY_PRUNE_DEPTH*DEPTH_INCREMENT) {
           const HistoryPruneParams &p = HISTORY_PRUNE_PARAMS[depth];
           // Late move pruning (LMP). Prune quiet moves that are late
           // in the move order (regardless of history score).
           if (moveIndex >= p.pruningMinMoveCount &&
               GetPhase(move) == MoveGenerator::HISTORY_PHASE) {
               return PRUNE;
           }
           else if (moveIndex >= p.historyMinMoveCount &&
                    History::scoreForPruning(move,board.sideToMove()) < p.evalThreshold) {
               // History pruning. Prune moves with low history scores.
               return PRUNE;
           }
       }
       // See pruning
       if (depth <= DEPTH_INCREMENT && 
           GetPhase(move) > MoveGenerator::WINNING_CAPTURE_PHASE &&
           see(board,move) < 0) {
          return PRUNE;
       }
    }
    // See if we do late move reduction. Moves in the history phase of move
    // generation or later can be searched with reduced depth.
    if (depth >= LMR_DEPTH && moveIndex > 1 &&
        ((ply == 0 && CanReduce(move)) ||
         GetPhase(move) >= MoveGenerator::HISTORY_PHASE) &&
        !passedPawnMove(board,move,6)) {
         // Moves that also meet the futility criteria are searched with greater reduction:
         extend -= LMR_REDUCTION[pruneOk][depth/DEPTH_INCREMENT][Util::Min(63,moveIndex)];
         node->extensions |= LMR;
    }
    return extend;
}

// Recursive function, implements alpha/beta search below ply 0 but
// above the quiescence search.
//
// The current board is this->board.
// The search parameters such as depth, ply are in this->node.
//
int Search::search()
{
    int rep_count;
    int using_tb;
    int move_count;
#ifdef MOVE_ORDER_STATS
    node->best_count = 0;
#endif
    int ply = node->ply;
    int depth = node->depth;
    if (time_check_counter++ >= Time_Check_Interval) {
        time_check_counter = 0;
        if (checkTime(board,ply)) {
            if (controller->talkLevel == Trace) {
                cout << "# terminating, time up" << endl;
            }
            controller->terminateNow();   // signal all searches to end
        }
    }
    ASSERT(ply < Constants::MaxPly);
    controller->stats->num_nodes++;
    nodeCount++;
    if (terminate) {
        return node->alpha;
    }
    else if (ply >= Constants::MaxPly-1) {
        return scoring.evalu8(board,node->alpha,node->beta);
    }

    if (Scoring::isDraw(board,rep_count,ply)) {
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "draw!" << endl;
        }
#endif
        node->flags |= EXACT;
        return drawScore(board);
    }
    Move hash_move = NullMove;
    using_tb = 0;
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
    int egtbDepth = Constants::MaxPly*DEPTH_INCREMENT;
    if (options.search.use_tablebases) {
        const Material &wMat = board.getMaterial(White);
        const Material &bMat = board.getMaterial(Black);
        egtbDepth = DEPTH_INCREMENT*root()->getIterationDepth()*3/4;
        using_tb = (wMat.men() + bMat.men() <= EGTBMenCount) && (depth >= egtbDepth || ply <= 2);
    }
#endif
    PositionInfo hashEntry;
    PositionInfo::ValueType result = PositionInfo::NoHit;
    // Note: query the hash table before the tablebases, since TB info may
    // be cached.
    if (options.search.hash_table_size) {
        // Search the hash table to see if we have hit this
        // position before.
        //
        // Note: we copy the hash entry .. so mods by another thread do not
        // alter the copy
        result = Hash::searchHash(board,board.hashCode(rep_count),
                            node->alpha,node->beta,ply,depth,controller->age,hashEntry);
#ifdef SEARCH_STATS
        controller->stats->hash_searches++;
#endif
    }
    bool hit = result != PositionInfo::NoHit;
    if (hit) {
#ifdef SEARCH_STATS
        controller->stats->hash_hits++;
#endif
         // always accept a full-depth entry (cached tb hit)
         if (depth != (Constants::MaxPly-1)*DEPTH_INCREMENT) {
            // if using TBs at this ply, do not pull a non-TB entry out of
            // the hashtable. Also do not allow hash based cutoff of a PV
            // node.
            if (using_tb || node->PV()) {
               result = PositionInfo::Invalid;
            }
        }
        int value = hashEntry.value();
        switch (result) {
            case PositionInfo::Valid:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = E" <<
                        " alpha = " << node->alpha <<
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                // If this is a mate score, adjust it to reflect the
                // current ply depth.
                //
                if (value >= Constants::MATE_RANGE) {
                    value -= ply;
                }
                else if (value <= -Constants::MATE_RANGE) {
                    value += ply;
                }
                if (node->inBounds(value)) {
                    // parent node will consider this a new best line
                    hash_move = hashEntry.bestMove(board);
                    if (!IsNull(hash_move)) {
                        node->pv[ply] = hash_move;
                        node->pv_length = 1;
                    }
#ifdef _TRACE
                    if (master()) {
                        indent(ply); cout << "best line[ply][ply] = ";
                        MoveImage(hash_move,cout);
                        cout << endl;
                    }
#endif
                }
                node->flags |= EXACT;
                return value;
            case PositionInfo::UpperBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = U" <<
                        " alpha = " << node->alpha <<
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                if (value <= node->alpha) {
                    return value;                     // cutoff
                }
                break;
            case PositionInfo::LowerBound:
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "hash cutoff, type = L" <<
                        " alpha = " << node->alpha <<
                        " beta = " << node->beta <<
                        " value = " << value << endl;
                }
#endif
                if (value >= node->beta) {
                    return value;                     // cutoff
                }
                break;
            default:
                break;
        } // end switch
        // Note: hash move & flags may be usable even if score is not usable
        hash_move = hashEntry.bestMove(board);
        // Entries from the learn file or from a TB hit do not have check status
        if (!hashEntry.tb() && !hashEntry.learned()) {
            board.setCheckStatus(hashEntry.inCheck() ? InCheck : NotInCheck);
        }
    }
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
    if (using_tb && rep_count==0) {
        int tb_hit = 0;
        controller->stats->tb_probes++;
        int tb_score;
#ifdef NALIMOV_TBS
         if (options.search.tablebase_type == Options::NALIMOV_TYPE) {
             tb_hit = NalimovTb::probe_tb(board, tb_score, ply);
         }
#endif
#ifdef GAVIOTA_TBS
         if (options.search.tablebase_type == Options::GAVIOTA_TYPE) {
             // TBD: use soft probing at lower depths
             tb_hit = GaviotaTb::probe_tb(board, tb_score, ply, true);
         }
#endif
        if (tb_hit) {
            controller->stats->tb_hits++;
#ifdef _TRACE
            if (master()) {
                indent(ply); cout << "EGTB hit: score " << tb_score << endl;
            }
#endif
            int score = tb_score;
            // insert TB info in hash table.
            if (score <= -Constants::MATE_RANGE) {
                score -= ply;
            }
            else if (score >= Constants::MATE_RANGE) {
                score += ply;
            }
#ifdef _TRACE
            if (master() && tb_score != score) {
                indent(ply); cout << "adjusted score " << score << endl;
            }
#endif
            // Put it in with a large depth so we will not
            // overwrite - this entry is "exact" at all
            // search depths, so effectively its depth is infinite.
            Hash::storeHash(board.hashCode(rep_count),
                (Constants::MaxPly-1)*DEPTH_INCREMENT,
                controller->age,
                PositionInfo::Valid,
                score,
                PositionInfo::FULL_MASK | PositionInfo::TB_MASK,
                NullMove);
            node->best_score = tb_score;               // unadjusted score
            node->flags |= EXACT;
            return node->best_score;
        }
    }
#endif
    // At this point we need to know if we are in check or not.
    int in_check =
        (board.checkStatus((node-1)->last_move) == InCheck);
#ifdef _DEBUG
    if (in_check != board.inCheck()) {
	   cout << "# " << board << endl;
	   cout << "# move=";
	   MoveImage((node-1)->last_move,cout);
	   cout << endl;
	   ASSERT(0);
    }
#endif
#ifdef _TRACE
    if (master() && in_check) { indent(ply); cout << "in_check=" << in_check << endl;}
#endif
    node->futilityEval = Scoring::INVALID_SCORE;
    node->fpruned_moves = 0;
    node->split = NULL;
    node->splits = 0;
    move_count = 0;

    bool doNull = !in_check &&
        !node->PV() &&
        !(node->flags & (IID|VERIFY)) &&
        !IsNull((node-1)->last_move) &&
        !Scoring::mateScore(node->alpha) &&
        board.state.moveCount <= 98 &&
        (board.getMaterial(board.sideToMove()).pieceCount() >= 1);

#ifdef STATIC_NULL_PRUNING
    // static null pruning, as in Stockfish, Protector, etc.
    if (doNull && depth <= 3*DEPTH_INCREMENT) {
        const int margin = FUTILITY_MARGIN_BASE[depth]/2;
        int threshold = node->beta+margin;
        int eval = scoring.evalu8(board,threshold-1,threshold);
        if (eval > threshold) {
            return eval - margin;
        }
    }
#endif

#ifdef RAZORING
    // razoring as in Glaurung & Toga
    if (doNull &&
        depth <= RAZOR_DEPTH && 
        depth>DEPTH_INCREMENT &&
        !(board.pawn_bits[board.sideToMove()] & Attacks::rank7mask[board.sideToMove()]) &&
         IsNull(hash_move)) {
      int threshold = node->beta - FUTILITY_MARGIN_BASE[(depth*4)/DEPTH_INCREMENT];
      int eval = node->futilityEval = scoring.evalu8(board,threshold-1,threshold);
      if (eval < threshold) {
         int v = quiesce(node->beta-1,node->beta,ply+1,0);
         if(v < node->beta) {
#ifdef _TRACE
           indent(ply); cout << "razored node, score=" << v << endl;
#endif
#ifdef SEARCH_STATS
           controller->stats->num_razored++;
#endif
           goto hash_insert;
         }
      }
    }
#endif

    // Try to get a fast cutoff using a "null move".  Skip if the side
    // to move is in check or if material balance is low enough that
    // zugzwang is a possibility. Do not do null move if this is an
    // IID search, because it will only help us get a cutoff, not a move.
    // Also avoid null move near the 50-move draw limit.
    node->threat = 0;
    node->threatMove = NullMove;
    if (doNull && depth >= 2*DEPTH_INCREMENT) {
        node->last_move = NullMove;
        BoardState state = board.state;
        board.doNull();
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "trying " << ply << ". " << "(null)" << endl;
        }
#endif
        int nu_depth;
        nu_depth = depth-3*DEPTH_INCREMENT-DEPTH_INCREMENT;
        // If score is very good do extra reduction (aka "smooth scaling").
        int threshold = node->beta + PAWN_VALUE;
        int eval;
        if (depth > 6*DEPTH_INCREMENT &&
            (eval = scoring.evalu8(board)) > threshold) {
            nu_depth -= Util::Min(DEPTH_INCREMENT,DEPTH_INCREMENT*(eval-node->beta-(PAWN_VALUE/2))/(2*PAWN_VALUE));
            nu_depth = Util::Max(nu_depth,DEPTH_INCREMENT);
        }
        int nscore;
        if (nu_depth > 0)
            nscore = -search(-node->beta, 1-node->beta,
                ply+1, nu_depth);
        else
            nscore = -quiesce(-node->beta, 1-node->beta,
                ply+1, 0);
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << ply << ". " << "(null)" << ' ' << nscore << endl;
        }
#endif
        board.undoNull(state);
        if (terminate) {
            node->best_score = node->alpha;
            goto search_end2;
        }
        else if (nscore >= node->beta) {          // cutoff
#ifdef VERIFY_NULL_SEARCH
          if (depth >= 6*DEPTH_INCREMENT) {
                // Verify null cutoff with reduced-depth search
                // (idea from Dieter Buerssner)
                nu_depth = depth-5*DEPTH_INCREMENT;
                ASSERT(nu_depth > 0);
                nscore = search(node->alpha, node->beta,
                    ply+1, nu_depth, VERIFY);
                if (nscore == -Illegal) {
#ifdef _TRACE
                    indent(ply); cout << "previous move illegal" << endl;
#endif
                    return -Illegal;
                }
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    if (nscore>=node->beta)
                        cout << "null cutoff verified, score=" << nscore;
                    else
                        cout << "null cutoff not verified";
                    cout << endl;
                }
#endif
            }
            if (nscore >= node->beta)             // null cutoff
#endif
            {
#ifdef _TRACE
                if (master()) {
                    indent(ply);
                    cout << "**CUTOFF**" << endl;
                }
#endif
#ifdef SEARCH_STATS
                controller->stats->num_null_cuts++;
#endif
                node->best_score = nscore;
                goto hash_insert;
            }
        }
        else {
           // null move failed low. Remember what the opponent move was
           // that refuted the null move.
           node->threatMove = (node+1)->best;
           // extend if the score was a mate
           if (nscore <= -Constants::MATE_RANGE) {
              node->threat++;
           }
        }
    }

    // Use "internal iterative deepening" to get an initial move to try if
    // there is no hash move .. an idea from Crafty (previously used by
    // Hitech).
    if (IsNull(hash_move) && depth >= 4*DEPTH_INCREMENT) {
        int d;
        d = Util::Min(depth-2*DEPTH_INCREMENT,depth/2);
        if (!node->PV()) d-=DEPTH_INCREMENT;
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "== deepening, depth = " << d
                << endl;
        }
#endif
        // Call search routine at lower depth to get a 1st move to try.
        // ("Internal iterative deepening").
        //
        // Note: we do not push down the node stack because we want this
        // search to have all the same parameters (including ply) as the
        // current search, just reduced depth + the IID flag set.
        node->flags |= IID;
        int alpha = node->alpha;
        node->num_try = 0;
        node->depth = d;
        int iid_score = -search();
        node->flags &= ~IID;
        // reset key params
        node->num_try = 0;
        node->cutoff = 0;
        node->depth = depth;
        node->alpha = node->best_score = alpha;
        node->splits = 0;
        node->fpruned_moves = 0;

        // set hash move to IID search result (may still be null)
        hash_move = node->best;
        // do not retain any pv information from the IID search
        // (can screw up non-IID pv).
        (node+1)->pv[ply+1] = NullMove;
        (node+1)->pv_length = 0;
        if (iid_score == Illegal || (node->flags & EXACT)) {
           // previous move was illegal or was an exact score
#ifdef _TRACE
          if (master()) {
             indent(ply);
             cout << "== exact result from deeping" << endl;
          }
#endif
           return -iid_score;
        }
        if (terminate) {
            node->best_score = node->alpha;
            return node->alpha;
        }
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "== deepening done.";
        }
#endif

#ifdef _TRACE
        if (master()) {
            if (!IsNull(hash_move)) {
                indent(ply); cout << "  hash_move = ";
                MoveImage(hash_move,cout);
            }
            cout << endl;
        }
#endif
    }
    {
        MoveGenerator mg(board, &context, ply, hash_move, master());
        node->mg = &mg;
        BoardState state = board.state;
        int try_score;

        // we do not split if in check because generally there will
        // be few moves to search there.
        const int canSplit = options.search.ncpus>1 && !in_check &&
          depth >= THREAD_SPLIT_DEPTH(board);
        //
        // Now we are ready to loop through the moves from this position
        //
#ifdef _TRACE
        int first = 1;
#endif
        int hibound = node->beta;
        while (!node->cutoff && !terminate) {
            Move move;
#ifdef _DEBUG
            ASSERT(node->splits == 0);
#endif
            // we do not bother to lock here, because if we have split
            // this node, we don't come back to this part of the loop
            move = in_check ? mg.nextEvasion() : mg.nextMove();
            if (IsNull(move)) break;
            if (IsUsed(move)) continue;
            if (Capture(move)==King) {
                return -Illegal;                  // previous move was illegal
            }
            move_count++;
#ifdef SEARCH_STATS
            controller->stats->num_moves++;
#endif
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << "trying " << ply << ". ";
                MoveImage(move,cout);
               if (first) cout << "(pv)";
                cout << " [" << node->best_score << "," << hibound << "]";
                cout << " d:" << depth << endl;
            }
#endif
            node->last_move = move;
            CheckStatusType in_check_after_move = board.wouldCheck(move);
            int extend = calcExtensions(board,node,node,in_check_after_move,
                                        node->num_try,move);
            if (extend == PRUNE) {
#ifdef _TRACE
              if (master()) {
                indent(ply); cout << "fwd pruned." << endl;
              }
#endif
              node->fpruned_moves++;
              continue;
            }
            board.doMove(move);
            if (!in_check && !board.wasLegal(move)) {
                  ASSERT(board.anyAttacks(board.kingSquare(board.oppositeSide()),board.sideToMove()));
#ifdef _TRACE
               if (master()) {
                  indent(ply); cout << "Illegal move!" << endl;
               }
#endif
               board.undoMove(move,state);
               continue;
            }
            setCheckStatus(board, in_check_after_move);
            if (depth+extend-DEPTH_INCREMENT > 0) {
                try_score = -search(-hibound, -node->best_score,
                    ply+1, depth+extend-DEPTH_INCREMENT);
            }
            else {
                try_score = -quiesce(-hibound, -node->best_score,
                    ply+1, 0);
            }
            if (try_score == Illegal) {
#if defined(_TRACE)
                if (master()) {
                    indent(ply); cout << "Illegal move" << endl;
                }
#endif
                board.undoMove(move,state);
                continue;
            }
            if (try_score > node->best_score && 
               (extend < 0 || hibound < node->beta) &&
                !((node+1)->flags & EXACT) &&
                !terminate) {
               // We failed to get a cutoff and must re-search
#ifdef _TRACE
               if (master()) {
                  indent(ply); cout << ply << ". ";
                  MoveImage(move,cout);
                  cout << " score = " << try_score << " - no cutoff, researching .." << endl;
                    indent(ply); cout << "window = [" << node->best_score << "," << node->beta
                        << "]" << endl;
               }
#endif
               if (extend < 0) {
                 extend = 0;
                 node->extensions = 0;
               }
               if (depth+extend-DEPTH_INCREMENT > 0)
                    try_score=-search(-node->beta,-node->best_score,ply+1,depth+extend-DEPTH_INCREMENT);
                else
                    try_score=-quiesce(-node->beta,-node->best_score,ply+1,0);
#ifdef _TRACE
               if (master()) {
                   indent(ply);
                   cout << ply << ". ";
                   MoveImage(move,cout);
                   cout << ' ' << try_score << endl;
               }
#endif
            }
            board.undoMove(move,state);
            if (terminate) {
                break;
            }
#ifdef _TRACE
            if (master()) {
                indent(ply);
                cout << ply << ". ";
                MoveImage(move,cout);
                cout << ' ' << try_score;
                if (first) cout << " (pv)";
                cout << endl;
            }
            first = 0;
#endif
            ASSERT(node->num_try<Constants::MaxMoves);
            node->done[node->num_try++] = move;
            ASSERT(node->num_try<Constants::MaxMoves);
            if (try_score > node->best_score) {
                if (updateMove(board,node,node,move,try_score,ply,depth)) {
                    // cutoff
                    break;
                }
            }
            if (try_score >= Constants::MATE-1-ply) {
                node->cutoff++;
                break;                            // mating move found
            }
            // zero-width window for the rest of the moves
            hibound = node->best_score+1;
            if (canSplit && maybeSplit(board,node,ply,depth)) {
                break;
            }
        }                                         // end move loop
#ifdef _TRACE
        if (node->best_score >= node->beta && master()) {
            indent(ply);
            cout << "**CUTOFF**" << endl;
        }
#endif
        // We may have other threads still searching.
        maybeWait(node);
        if (terminate) {
            node->best_score = node->alpha;
            goto search_end2;
        }
        if (node->num_try == 0 && node->fpruned_moves == 0
        ) {
            // no moves were tried
            if (in_check) {
#ifdef _TRACE
                if (master()) {
                    indent(ply); cout << "mate" << endl;
                }
#endif
                node->best_score = -(Constants::MATE - ply);
                node->flags |= EXACT;
                goto search_end2;
            }
            else {                                // stalemate
#ifdef _TRACE
                if (master()) {
                    indent(ply); cout << "stalemate!" << endl;
                }
#endif
                node->best_score = drawScore(board);
                node->flags |= EXACT;
                return node->best_score;
            }
        }
    }
    if (node->num_try && (node->best_score >= node->beta) &&
        TypeOfMove(node->best) == Normal && Capture(node->best) == Empty) {
        History::updateHistory(board,node,node->best,
            depth,
            board.sideToMove());
    }

    // don't insert into the hash table if we are terminating - we may
    // not have an accurate score.
 hash_insert:
    if (options.search.hash_table_size && !terminate) {
        if (node->best == NullMove) node->best = hash_move;
        // store the position in the hash table, if there's room
        int value = node->best_score;
        PositionInfo::ValueType val_type;
        // Adjust mate scores to reflect current ply. But only
        // if the score is in bounds.
        if (value > node->alpha && value < node->beta) {
            if (value < -Constants::MATE_RANGE) {
                value -= ply;
            }
            else if (value > Constants::MATE_RANGE) {
                value += ply;
            }
        }
#ifdef _TRACE
        char typeChar;
#endif
        if (value <= node->alpha) {
            val_type = PositionInfo::UpperBound;
#ifdef _TRACE
            typeChar = 'U';
#endif
            // We don't have a "best" move, because all moves
            // caused alpha cutoff.  But if there was a hash
            // move or an initial move produced by internal
            // interative deepening, save it in the hash table
            // so it will be tried first next time around.
            node->best = hash_move;
        }
        else if (value >= node->beta) {
            val_type = PositionInfo::LowerBound;
#ifdef _TRACE
            typeChar = 'L';
#endif
        }
        else {
            val_type = node->fpruned_moves ? PositionInfo::Invalid : PositionInfo::Valid;
#ifdef _TRACE
            typeChar = 'E';
#endif
        }
#ifdef _TRACE
        if (master()) {
            indent(ply);
            cout << "storing type=" << typeChar <<
                " ply=" << ply << " depth=" << depth << " value=" << value <<
                " move=";
            MoveImage(node->best,cout);
            cout << endl;
        }
#endif
        hash_t hash = board.hashCode(rep_count);
        Hash::storeHash(hash, depth,
            controller->age,
            val_type,
            node->best_score, 
            (node->PV() ? PositionInfo::FULL_MASK : 0) |
            (board.checkStatus() == InCheck ? PositionInfo::CHECK_MASK : 0) |
            (IsForced(node->best) ? PositionInfo::FORCED_MASK : 0) |
            (IsForced2(node->best) ? PositionInfo::FORCED2_MASK : 0),
            node->best);
    }
    search_end2:
#ifdef MOVE_ORDER_STATS
    if (node->num_try && node->best_score != initial_alpha) {
        controller->stats->move_order_count++;
        ASSERT(node->best_count>=0);
        if (node->best_count<4) {
            controller->stats->move_order[node->best_count]++;
        }
    }
#endif
#ifdef SEARCH_STATS
    controller->stats->futility_pruning += node->fpruned_moves;
#endif
    int score = node->best_score;
    ASSERT(score >= -Constants::MATE && score <= Constants::MATE);
    return score;
}

int Search::updateRootMove(const Board &board,
NodeInfo *parentNode, NodeInfo *node,
Move move, int score)
{
   int locked = 0;
   if (parentNode->split) {
       parentNode->lock();
       locked++;
   }
   if (score > parentNode->best_score)  {
      parentNode->best = move;
      ASSERT(parentNode->best_score >=-Constants::MATE && parentNode->best_score <=Constants::MATE);
      if (!CaptureOrPromotion(node->last_move)) {
         context.setKiller(node->last_move, 0);
      }
      parentNode->best_score = score;
#ifdef MOVE_ORDER_STATS
      parentNode->best_count = node->num_try-1;
#endif
      if (score >= parentNode->beta) {
#ifdef _TRACE
         if (master())
             cout << "ply 0 beta cutoff" << endl;
#endif
         // set pv to this move so it is searched first the next time
         parentNode->pv[0] = move;
         parentNode->pv_length = 1;
         parentNode->cutoff++;
         parentNode->best_score = score;
         controller->updateStats(parentNode, controller->getIterationDepth(),
                            parentNode->best_score,
                            parentNode->alpha,parentNode->beta);
         showStatus(board,move,score <= parentNode->alpha,score >= parentNode->beta,0);
         if (locked) {
            parentNode->unlock();
         }
         if (controller->uci) {
            cout << "info score ";
            Scoring::printScoreUCI(score,cout);
            cout << " lowerbound" << endl;
         }
         return 1;  // signal cutoff
      }
      updatePV(board,parentNode,(node+1),move,0);
      controller->updateStats(parentNode, controller->getIterationDepth(),
                              parentNode->best_score,
                              parentNode->alpha,parentNode->beta);
      showStatus(board,move,score <= parentNode->alpha,score >= parentNode->beta,0);
   }
   if (locked) {
      parentNode->unlock();
   }
   return 0;   // no cutoff
}

void Search::searchSMP(ThreadInfo *ti)
{
    Move move;
    const int in_check = board.checkStatus() == InCheck;
    BoardState state(board.state);
    const int &ply = node->ply;
    const int &depth = node->depth;
    // Count moves from the point where the parent was split.
    // This is used for pruning and reduction. Note this will
    // generally under-estimate the actual number of moves 
    // that have been searched when multiple threads are at work,
    // but that's ok: we would rather under-prune than prune
    // excessively.
    int moveIndex = node->num_try;
    NodeInfo *parentNode = ti->work->node->parentNode;
    //    parentNode->lock();
    int best_score = parentNode->best_score;
    //    parentNode->unlock();
#ifdef _THREAD_TRACE
    log("searchSMP",ti->index);
#endif
    while (!parentNode->cutoff && !terminate) {
        move = in_check ?
            parentNode->mg->nextEvasion(parentNode) :
            parentNode->mg->nextMove(parentNode);
        if (IsNull(move)) break;
        if (IsUsed(move)) continue;
#ifdef _TRACE
        if (master() || ply==0) {
            indent(ply); cout << "trying " << ply << ". ";
            MoveImage(move,cout); cout << endl;
        }
#endif
        ASSERT(Capture(move) != King);
        int try_score;

        node->extensions = 0;
        node->last_move = move;
        CheckStatusType in_check_after_move = board.wouldCheck(move);
        int extend = calcExtensions(board,node,parentNode,in_check_after_move,
                                    moveIndex,move);
        if (extend == PRUNE) {
#ifdef _TRACE
           if (master()) {
              indent(ply); cout << "fwd pruned." << endl;
           }
#endif
           parentNode->fpruned_moves++;
           continue;
        }
        board.doMove(move);
        if (!in_check && !board.wasLegal(move)) {
#ifdef _TRACE
           if (master()) {
               indent(ply); cout << "Illegal move!" << endl;
           }
#endif
           board.undoMove(move,state);
           continue;
        }
        setCheckStatus(board,in_check_after_move);
#ifdef _TRACE
        if (master()) {
            indent(ply); cout << "window [" << best_score <<
                "," << best_score+1 << "]" << endl;
        }
#endif
        if (depth+extend-DEPTH_INCREMENT > 0) {
            try_score = -search(-best_score-1, -best_score,
                ply+1, depth+extend-DEPTH_INCREMENT);
        }
        else {
            try_score = -quiesce(-best_score-1, -best_score,
                ply+1, 0);
        }
        if (try_score == Illegal) {
#ifdef _TRACE
            if (master()) {
                indent(ply); cout << "Illegal move" << endl;
            }
#endif
            board.undoMove(move,state);
            continue;
        }
        // keep count in parent of legal moves searched
        parentNode->num_try++;
        moveIndex++;
#ifdef _TRACE
        if (master() || ply==0) {
            indent(ply); cout << ply << ". ";
            MoveImage(move,cout);
            cout << " " << try_score << endl;
        }
#endif
        if (terminate || parentNode->cutoff) {
#ifdef _TRACE
          if (master()) {
             indent(ply); cout << "parent node cutoff" << endl;
          }
#endif
           board.undoMove(move,state);
           break;
        }
        // it is possible the parent node's best score has changed, so
        // compare against that
        parentNode->lock();
        if (try_score > parentNode->best_score && 
            (parentNode->beta > best_score+1 || extend < 0) &&
            !((node+1)->flags & EXACT) &&
            !terminate) {
            // We got a new best move but with a zero-width search or with
            // reduction enabled, so we must re-search.
#ifdef _TRACE
            if (master() || ply==0) {
                indent(ply); cout << ply << ". ";
                MoveImage(move,cout);
                cout << "score = " << try_score << " - no cutoff, researching";
                cout << " (smp ";
                cout << ti->index << ')' << endl;
            }
#endif
            parentNode->unlock();
            if (extend < 0) {
                node->extensions = 0;
                extend = 0;
            }
            if (depth+extend-DEPTH_INCREMENT > 0)
                try_score=-search(-parentNode->beta,-best_score,ply+1,depth+extend-DEPTH_INCREMENT);
            else
                try_score=-quiesce(-parentNode->beta,-best_score,ply+1,0);
        }
        else
           parentNode->unlock();
        board.undoMove(move,state);
#ifdef _TRACE
        if (master() || ply==0) {
            indent(ply);
            cout << ply << ". ";
            MoveImage(move,cout);
            cout << " " << try_score << endl;
        }
#endif
        if (!terminate && !parentNode->cutoff) {
            parentNode->lock();
            ASSERT(parentNode->num_try<Constants::MaxMoves);
            parentNode->done[parentNode->num_try++] = move;
            parentNode->unlock();
            if (try_score > parentNode->best_score) {
                // search produced a new best move or cutoff, update parent node
#if defined (_THREAD_TRACE) || defined(_TRACE)
              stringstream s;
                s << "new best smp, thread=" << ti->index << " ply= " << ply << '\0';
#ifdef _TRACE
                cout << s.str() << endl;
#endif
#ifdef _THREAD_TRACE
                log(s.str());
#endif
#endif
                if (ply == 0)
                    ((RootSearch*)(ti->work->parent))->updateRootMove(board,parentNode,node,move,try_score);
                else
                    ti->work->parent->updateMove(board,parentNode,node,move,try_score,ply,depth);
                best_score = parentNode->best_score;
            }
            if (try_score >= Constants::MATE-1-ply) {
                parentNode->cutoff++;
                break;                            // mating move found
            }
        }
        if (ply == 0 && waitTime) {
            sleep(waitTime);
        }
    }
    // this search is done, make this thread available for
    // other tasks
}

int Search::updateMove(const Board &board, NodeInfo *parentNode, NodeInfo *node, Move move, int score, int ply, int depth)
{
    int cutoff = 0;
    int locked = 0;
    if (parentNode->split) {
       parentNode->lock();
       locked++;
    }
    // avoid race - check score again when lock acquired
    if (score > parentNode->best_score) {
       parentNode->best_score = score;
       parentNode->best = move;
       if (Capture(move) == Empty && PromoteTo(move) == Empty) {
           if (node->best_score >= parentNode->beta) {
               context.setKiller(move, ply);
           }
       }
#ifdef MOVE_ORDER_STATS
       parentNode->best_count = node->num_try-1;
#endif
       if (score >= parentNode->beta) {
#ifdef _TRACE
           if (master()) {
               indent(ply); cout << "beta cutoff" << endl;
           }
#endif
           parentNode->cutoff++;
           cutoff++;
       }
       else {
           parentNode->best_score = score;
           // update pv from slave node to master
           updatePV(board,parentNode,(node+1),move,ply);
       }
    }
    if (locked) parentNode->unlock();
    return cutoff;
}

void Search::updatePV(const Board &board, Move m, int ply)
{
    updatePV(board,node,(node+1),m,ply);
#ifdef _DEBUG
    Board board_copy(board);
#ifdef _TRACE
    if (master()) {
        indent(ply); cout << "update_pv, ply " << ply << endl;
    }
#endif
    for (int i = ply; i < node->pv_length+ply; i++) {
#ifdef _TRACE
        if (master()) {
            MoveImage(node->pv[i],cout); cout << " " << (flush);
        }
#endif
        ASSERT(legalMove(board_copy,node->pv[i]));
        board_copy.doMove(node->pv[i]);
    }
#endif
#ifdef _TRACE
    if (master()) cout << endl;
#endif
}


void Search::updatePV(const Board &board,NodeInfo *node,NodeInfo *fromNode,Move move, int ply)
{
#ifdef _TRACE
    if (master()) {
        indent(ply); cout << "update_pv, ply " << ply << endl;
    }
#endif
    node->pv[ply] = move;
    if (fromNode->pv_length) {
        memcpy((void*)(node->pv+ply+1),(void*)(fromNode->pv+ply+1),
            sizeof(Move)*fromNode->pv_length);
    }
    node->pv_length = fromNode->pv_length+1;
#ifdef _DEBUG
    Board board_copy(board);
    for (int i = ply; i < node->pv_length+ply; i++) {
                ASSERT(i<Constants::MaxPly);
#ifdef _TRACE
        if (master()) {
            MoveImage(node->pv[i],cout); cout << " " << (flush);
        }
#endif
        ASSERT(legalMove(board_copy,node->pv[i]));
        board_copy.doMove(node->pv[i]);
    }
#endif
}


// stop all threads associated with this search
void Search::stopAllThreads()
{
#ifdef _THREAD_TRACE
    log("waiting, discarding results");
#endif
    terminate = 1;
    if (options.search.ncpus>1) {
        Lock(splitLock);
#ifdef _THREAD_TRACE
        log("stopAllThreads",ti->index);
#endif
        split_t *s = splitStackEnd;
        ASSERT(splitStackEnd-splitStack<=SPLIT_STACK_MAX_DEPTH);
        while (s != splitStack) {
            --s;
            Lock(s->nodeLock);
            const int max = s->slaves.size();
            for (int i = 0; i<max; i++) {
                (s->slaves[i])->work->stopAllThreads();
            }
            Unlock(s->nodeLock);
        }
        Unlock(splitLock);
    }
}

// Initialize the saved state from the current Search instance
void Search::SearchState::init(const Search *source) {
  ti = source->ti;
  parent = (Search*)(source->parent);
  board = source->board;
  nodeContents = (*(source->node));
  node = (source->node);
  splitStackEnd = source->splitStackEnd;
}

void Search::restore(Search::SearchState &state) {
  ti = state.ti;
  parent = state.parent;
  board = state.board;
  node = state.node;
  *node = state.nodeContents;
  splitStackEnd = state.splitStackEnd;
}

split_t *Search::split(NodeInfo *nodeToSplit,ThreadInfo *ti,int ply,int depth)
{
    Lock(splitLock);
    if (!nodeToSplit->split) {
        // This node didn't have a split point already, so make one,
        // if possible:
        ASSERT(parent->splitStackEnd>=parent->splitStack);
        if (splitStackEnd-splitStack >= SPLIT_STACK_MAX_DEPTH) {
            Unlock(splitLock);
            return NULL;   // can't split this search instance any more
        }
        ASSERT(splitStack);
        nodeToSplit->split = splitStackEnd++;
        ASSERT(nodeToSplit->split->slaves.size()==0);
    }
    ASSERT(nodeToSplit->split);
    nodeToSplit->lock();
    nodeToSplit->split->ply = ply;
    nodeToSplit->split->depth = depth;
    nodeToSplit->split->slaves.add(ti);
    nodeToSplit->splits++;
    ti->parentNode = nodeToSplit;
#ifndef _WIN32
    // ensure parent thread will wait
    parent->ti->reset();
#endif
    nodeToSplit->unlock();
    Unlock(splitLock);
    return nodeToSplit->split;
}


// return # of still active threads
int Search::join(NodeInfo *splitNode,ThreadInfo *ti)
{
#ifdef _THREAD_TRACE
    log("join",ti->index);
#endif
    Search *worker = ti->work;
    // update the node count for the root move we were searching
    (*ply0nodeCount) += worker->nodeCount;
    Search *parent = worker->parent;
    Lock(parent->splitLock);
    splitNode->lock();
    // dissociate the thread from the parent
    ASSERT(splitNode->split);
    ArasanSet<ThreadInfo *,Constants::MaxCPUs> &slaves =
        splitNode->split->slaves;
    // remove ti from the list of slave threads in the parent
#ifdef _DEBUG
    ASSERT(slaves.remove(ti));
#else
    slaves.remove(ti);
#endif
    ASSERT(slaves.size()>=0 &&
                slaves.size()<Constants::MaxCPUs);
    int remaining = slaves.size();
    splitNode->unlock();
    Unlock(parent->splitLock);
    return remaining;
}

int Search::maybeSplit(const Board &board, NodeInfo *node,int ply,int depth)
{
    // Now that we have searched at least one valid move, we can
    // consider using multiple threads to search the rest (YBWC).
    int splits = 0;
    if (!terminate && node->mg->more() &&
#if defined(GAVIOTA_TBS) || defined(NALIMOV_TBS)
        (!options.search.use_tablebases ||
        board.getMaterial(White).men() +
        board.getMaterial(Black).men() > EGTBMenCount) &&
#endif
        controller->pool->checkAvailable()) {

        ThreadInfo *child_ti;
        int remaining = 100;
        ThreadInfo *slaves[Constants::MaxCPUs];
        int moveIndex = node->num_try;
        // Keep trying to get/assign threads until no threads are
        // available or no more moves are available to search.
        while (remaining>1 && (child_ti = controller->pool->checkOut(this,node,ply,depth))!=NULL) {
            // A thread is available.
            if (!splits) {
                // Force all remaining moves to be generated, since the
                // MoveGenerator class is not thread-safe otherwise (it
                // maintains a pointer to the board and when accessed by
                // multiple threads this pointer may not always be at the
                // current position). Also we want to know how many moves
                // remain.
                remaining = node->mg->generateAllMoves(node);
            }
            --remaining;
            slaves[ splits++] = child_ti;
#ifdef SMP_STATS
            controller->stats->splits++;
#endif
            // A thread is available, copy state so the slave thread can
            // run independently of its master:
            child_ti->work->init(board,this,node,child_ti,ply,depth,moveIndex++,waitTime);
#ifdef _THREAD_TRACE
            log("split ply",ply);
            log("depth",depth);
#endif
        }
        // now start the slave threads
	Lock(splitLock);
        for (int i = 0; i < splits; i++) {
            // Start searching in the new thread:
            slaves[i]->start();
        }
	Unlock(splitLock);
    }
    // Now the master thread is still available to do work, either
    // on its own or on behalf of a slave thread ("helpful master").
    // Enter the idleLoop function, which will continue searching
    // until all work is complete.
    if (splits) idleLoop(node);
    return splits;
}


void Search::maybeWait(NodeInfo *node)
{
    if (node->wasSplit()) {
        if (node->hasSplit()) {
            if (node->cutoff || terminate) {
                // We just need to stop the other threads.
                // We don't care about their results.
                stopAllThreads(node);
            }
            waitForThreadCompletion(node);
        }
        // there are now no slave threads, so pop the split point stack
        Lock(splitLock);
        ASSERT(node->split->slaves.size()==0);
        node->split = NULL;
        node->splits = 0;
        splitStackEnd--;
        ASSERT(splitStackEnd>=splitStack);
        Unlock(splitLock);
        // ensure this thread will wait once back in the idle loop
#ifndef _WIN32
        ti->reset();
#endif
    }
}


// initialize a Search instance to prepare it for searching at a split point
void Search::init(const Board &aboard, Search *p, NodeInfo *parentNode, ThreadInfo *child_ti, int ply, int depth, int moveIndex, int timeToWait)
{
    parent = p;
    board = aboard;
    waitTime = timeToWait;
    // clear split stack
    if (this->ti && this->ti->master)
       splitStackEnd = saved.splitStackEnd;
    else
       splitStackEnd = splitStack;
    // inherit the parent node count so we can keep track of how many
    // nodes have been searched at ply 0 in all threads (for move
    // ordering)
    ply0nodeCount = p->ply0nodeCount;
    nodeCount = 0ULL; // count of nodes searched in this thread
    this->ti = child_ti;
    terminate = p->terminate;
    // set node ptr to our position in the stack
    node = stack+ply;
    // Copy a little info from the parent node to the child.
    // Note: we do not copy all of it, since the child will
    // mostly access the parent node for state.
    node->ply = ply;
    node->depth = depth;
    node->num_try = moveIndex;
    node->beta = (p->stack+ply)->beta;
    if (ply>0) {
      // There is some info from the previous node in the stack
      // that the search needs to access, so copy it from the parent too:
      (node-1)->extensions = (p->stack+ply-1)->extensions;
      (node-1)->last_move = (p->stack+ply-1)->last_move;
    }
    // set back pointer to parent
    node->parentNode = parentNode;
    // clear killer since the side to move may have been different
    // in the previous use of this class.
    context.clearKiller();
}


// stop all threads at or below node n
void Search::stopAllThreads(NodeInfo *n)
{
    Lock(splitLock);
    if (n->split == NULL) {
        Unlock(splitLock);
        return;  // no slave threads
    }
    split_t *s = (split_t*)splitStackEnd;
    // in case of many threads, there may be multiple split points
    // at various depths in the current search
    split_t *top = (split_t*)n->split;
    while (s != top) {
        --s;
        ASSERT(s>=splitStack);
	Lock(s->nodeLock);
        const int max = s->slaves.size();
        for (int i = 0; i<max; i++) {
            // this call has no parameter because all searching in the
            // associated Search class is to stop
            (s->slaves[i])->work->stopAllThreads();
        }
	Unlock(s->nodeLock);
    }
    Unlock(splitLock);
}


void Search::waitForThreadCompletion(NodeInfo *n)
{
    if (options.search.ncpus<=1) return;
    while (n->hasSplit()) {
#ifdef _TRACE
        indent(n->ply);cout << "waiting .. " << endl;
#endif
        // this node has an active split point
#ifdef _THREAD_TRACE
        log("waiting",ti->index);
#endif
        if (ti->wait()) break;
    }
#ifdef _THREAD_TRACE
    log("out of wait",ti->index);
#endif
}

void Search::idleLoop(NodeInfo *node)
{
    // This node is the master, so it is its own parent (score updates, pv etc.
    // are kept in the parentNode)
    node->parentNode = node;
    // Go ahead and try to grab/search moves in this thread, in parallel
    // with its slaves:
    searchSMP(ti);
}

void Search::clearHashTables() {
  node = stack;
  parent = this;
  scoring.clearHashTables();
}



